@@ -5,6 +5,7 @@ import inspect
 import ast
 import importlib
 import warnings
+import time
 from pytools import memoize
 from .config import get_config
 from .cython_generator import CythonGenerator
@@ -13,6 +14,7 @@ from .types import (dtype_to_ctype, get_declare_info,
                     dtype_to_knowntype, annotate)
 from .extern import Extern
 from .utils import getsourcelines
+from .profile import profile
 
 from . import array
 from . import parallel
@@ -32,31 +34,6 @@ def memoize_kernel(key=lambda *args: args):
     return memoize_deco
 
 
-def kernel_cache_key_args(obj, *args):
-    key = [get_ctype_from_arg(arg) for arg in args]
-    key.append(obj.func)
-    key.append(obj.name)
-    key.append(obj.backend)
-    key.append(obj._config.use_openmp)
-    return tuple(key)
-
-
-def kernel_cache_key_kwargs(obj, **kwargs):
-    key = [get_ctype_from_arg(arg) for arg in kwargs.values()]
-    key.append(obj.input_func)
-    key.append(obj.output_func)
-    key.append(obj.scan_expr)
-    key.append(obj.backend)
-    key.append(obj._config.use_openmp)
-    return tuple(key)
-
-
-def getargspec(f):
-    getargspec_f = getattr(inspect, 'getfullargspec',
-                           getattr(inspect, 'getargspec'))
-    return getargspec_f(f)[0]
-
-
 def get_ctype_from_arg(arg):
     if isinstance(arg, array.Array):
         return arg.gptr_type
@@ -72,6 +49,27 @@ def get_ctype_from_arg(arg):
                 return 'int'
 
 
+def kernel_cache_key_args(obj, *args):
+    key = [get_ctype_from_arg(arg) for arg in args]
+    key.append(obj.func)
+    key.append(obj.name)
+    return tuple(key + list(parallel.get_common_cache_key(obj)))
+
+
+def kernel_cache_key_kwargs(obj, **kwargs):
+    key = [get_ctype_from_arg(arg) for arg in kwargs.values()]
+    key.append(obj.input_func)
+    key.append(obj.output_func)
+    key.append(obj.scan_expr)
+    return tuple(key + list(parallel.get_common_cache_key(obj)))
+
+
+def getargspec(f):
+    getargspec_f = getattr(inspect, 'getfullargspec',
+                           getattr(inspect, 'getargspec'))
+    return getargspec_f(f)[0]
+
+
 def get_binop_return_type(a, b):
     if a is None or b is None:
         return None
@@ -167,7 +165,7 @@ class AnnotationHelper(ast.NodeVisitor):
         # make this work
         mod = importlib.import_module(self.func.__module__)
         f = getattr(mod, node.func.id, None)
-        if not hasattr(f, 'is_jit'):
+        if node.func.id not in BUILTINS and not hasattr(f, 'is_jit'):
             return None
         if node.func.id in self.external_funcs:
             return self.external_funcs[node.func.id].arg_types.get(
@@ -280,9 +278,7 @@ class AnnotationHelper(ast.NodeVisitor):
             else:
                 self.arg_types['return_'] = 'double'
         else:
-            if not node.value:
-                self.warn("Not returning anything", node)
-            else:
+            if node.value:
                 self.warn("Unknown type for return value. "
                           "Return value defaulting to 'double'", node)
                 self.arg_types['return_'] = 'double'
@@ -297,6 +293,8 @@ class ElementwiseJIT(parallel.ElementwiseBase):
         self.func = func
         self._config = get_config()
         self.cython_gen = CythonGenerator()
+        self.source = '# Code jitted, call the function to generate the code.'
+        self.all_source = self.source
         if backend == 'opencl':
             from .opencl import get_context, get_queue
             self.queue = get_queue()
@@ -331,6 +329,7 @@ class ElementwiseJIT(parallel.ElementwiseBase):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, *args, **kw):
         c_func = self._generate_kernel(*args)
         c_args = [self._massage_arg(x) for x in args]
@@ -372,6 +371,8 @@ class ReductionJIT(parallel.ReductionBase):
             self.neutral = neutral
         self._config = get_config()
         self.cython_gen = CythonGenerator()
+        self.source = '# Code jitted, call the function to generate the code.'
+        self.all_source = self.source
         if backend == 'opencl':
             from .opencl import get_context, get_queue
             self.queue = get_queue()
@@ -406,6 +407,7 @@ class ReductionJIT(parallel.ReductionBase):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, *args, **kw):
         c_func = self._generate_kernel(*args)
         c_args = [self._massage_arg(x) for x in args]
@@ -452,6 +454,8 @@ class ScanJIT(parallel.ScanBase):
         else:
             self.neutral = neutral
         self._config = get_config()
+        self.source = '# Code jitted, call the function to generate the code.'
+        self.all_source = self.source
         self.cython_gen = CythonGenerator()
         if backend == 'opencl':
             from .opencl import get_context, get_queue
@@ -513,10 +517,17 @@ class ScanJIT(parallel.ScanBase):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, **kwargs):
         c_func = self._generate_kernel(**kwargs)
         c_args_dict = {k: self._massage_arg(x) for k, x in kwargs.items()}
-        output_arg_keys = self.output_func.arg_keys[self._get_backend_key()]
+        if self._get_backend_key() in self.output_func.arg_keys:
+            output_arg_keys = self.output_func.arg_keys[
+                    self._get_backend_key()]
+        else:
+            raise ValueError("No kernel arguments found for backend = %s, "
+                             "use_openmp = %s, use_double = %s" %
+                             self._get_backend_key())
 
         if self.backend == 'cython':
             size = len(c_args_dict[output_arg_keys[1]])
