@@ -13,6 +13,7 @@ from mako.template import Template
 import numpy as np
 
 from .config import get_config
+from .profile import profile
 from .cython_generator import get_parallel_range, CythonGenerator
 from .transpiler import Transpiler, convert_to_float_if_needed
 from .types import dtype_to_ctype
@@ -396,6 +397,10 @@ def serial(func=None, **kw):
         return wrapper(func)
 
 
+def get_common_cache_key(obj):
+    return obj.backend, obj._config.use_openmp, obj._config.use_double
+
+
 class ElementwiseBase(object):
     def __init__(self, func, backend=None):
         backend = array.get_backend(backend)
@@ -406,6 +411,10 @@ class ElementwiseBase(object):
         self._config = get_config()
         self.cython_gen = CythonGenerator()
         self.queue = None
+        # This is the source generated for the user code.
+        self.source = '# Source not yet generated.'
+        # This is all the source code used for the elementwise.
+        self.all_source = '# Source not yet generated.'
         self.c_func = self._generate()
 
     def _generate(self, declarations=None):
@@ -427,8 +436,12 @@ class ElementwiseBase(object):
                     self.func, 'is_serial', False),
                 get_parallel_range=get_parallel_range
             )
+            # This is the user code source.
+            self.source = self.tp.get_code()
             self.tp.add_code(src)
             self.tp.compile()
+            # All the source code for the elementwise
+            self.all_source = self.tp.source
             return getattr(self.tp.mod, 'py_' + self.name[7:])
         elif self.backend == 'opencl':
             py_data, c_data = self.cython_gen.get_func_signature(self.func)
@@ -456,6 +469,10 @@ class ElementwiseBase(object):
                 operation=expr,
                 preamble="\n".join([cluda_preamble, preamble])
             )
+            # only code we generate is saved here.
+            self.source = "\n".join([cluda_preamble, preamble])
+            all_source = knl.get_kernel(False)[0].program.source
+            self.all_source = all_source or self.source
             return knl
         elif self.backend == 'cuda':
             py_data, c_data = self.cython_gen.get_func_signature(self.func)
@@ -481,6 +498,10 @@ class ElementwiseBase(object):
                 operation=expr,
                 preamble="\n".join([cluda_preamble, preamble])
             )
+            # only code we generate is saved here.
+            self.source = cluda_preamble + preamble
+            # FIXME: it is difficult to get the sources from pycuda.
+            self.all_source = self.source
             return knl
 
     def _correct_opencl_address_space(self, c_data):
@@ -515,6 +536,7 @@ class ElementwiseBase(object):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, *args, **kw):
         c_args = [self._massage_arg(x) for x in args]
         if self.backend == 'cython':
@@ -544,6 +566,9 @@ class Elementwise(object):
     def __getattr__(self, name):
         return getattr(self.elementwise, name)
 
+    def __dir__(self):
+        return sorted(dir(self.elementwise) + ['elementwise'])
+
     def __call__(self, *args, **kwargs):
         self.elementwise(*args, **kwargs)
 
@@ -581,6 +606,10 @@ class ReductionBase(object):
         self._config = get_config()
         self.cython_gen = CythonGenerator()
         self.queue = None
+        # This is the source generated for the user code.
+        self.source = '# Source not yet generated.'
+        # This is all the source code used.
+        self.all_source = '# Source not yet generated.'
         self.c_func = self._generate()
 
     def _generate(self, declarations=None):
@@ -614,8 +643,11 @@ class ReductionBase(object):
                 openmp=self._config.use_openmp,
                 get_parallel_range=get_parallel_range
             )
+            # This is the user code source.
+            self.source = self.tp.get_code()
             self.tp.add_code(src)
             self.tp.compile()
+            self.all_source = self.tp.source
             return getattr(self.tp.mod, 'py_' + self.name)
         elif self.backend == 'opencl':
             if self.func is not None:
@@ -654,6 +686,17 @@ class ReductionBase(object):
                 arguments=arguments,
                 preamble="\n".join([cluda_preamble, preamble])
             )
+            # only code we generate is saved here.
+            self.source = "\n".join([cluda_preamble, preamble])
+            if knl.stage_1_inf.source:
+                self.all_source = "\n".join([
+                    "// ------ stage 1 -----",
+                    knl.stage_1_inf.source,
+                    "// ------ stage 2 -----",
+                    knl.stage_2_inf.source,
+                ])
+            else:
+                self.all_source = self.source
             return knl
         elif self.backend == 'cuda':
             if self.func is not None:
@@ -690,6 +733,10 @@ class ReductionBase(object):
                 arguments=arguments,
                 preamble="\n".join([cluda_preamble, preamble])
             )
+            # only code we generate is saved here.
+            self.source = cluda_preamble + preamble
+            # FIXME: it is difficult to get the sources from pycuda.
+            self.all_source = self.source
             return knl
 
     def _correct_return_type(self, c_data):
@@ -737,6 +784,7 @@ class ReductionBase(object):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, *args):
         c_args = [self._massage_arg(x) for x in args]
         if self.backend == 'cython':
@@ -772,6 +820,9 @@ class Reduction(object):
                                           neutral=neutral,
                                           backend=backend)
 
+    def __dir__(self):
+        return sorted(dir(self.reduction) + ['reduction'])
+
     def __getattr__(self, name):
         return getattr(self.reduction, name)
 
@@ -804,12 +855,16 @@ class ScanBase(object):
         else:
             self.neutral = neutral
         self._config = get_config()
+        # This is the source generated for the user code.
+        self.source = '# Source not yet generated.'
+        # This is all the source code used for the elementwise.
+        self.all_source = '# Source not yet generated.'
         self.cython_gen = CythonGenerator()
         self.queue = None
         self.c_func = self._generate()
 
     def _get_backend_key(self):
-        return (self.backend, self._config.use_openmp, self._config.use_double)
+        return get_common_cache_key(self)
 
     def _correct_return_type(self, c_data, modifier):
         code = self.tp.blocks[-1].code.splitlines()
@@ -835,19 +890,12 @@ class ScanBase(object):
         else:
             return False
 
-    def _ignore_arg(self, arg_name):
-        if arg_name in ['item', 'prev_item', 'last_item', 'i', 'N']:
-            return True
-        return False
+    def _not_ignored(self, args):
+        ignore = ['item', 'prev_item', 'last_item', 'i', 'N']
+        return [i for (i, x) in enumerate(args) if x not in ignore]
 
-    def _num_ignore_args(self, c_data):
-        result = 0
-        for arg_name in c_data[1][:]:
-            if self._ignore_arg(arg_name):
-                result += 1
-            else:
-                break
-        return result
+    def _filter_ignored(self, args, indices):
+        return [args[x] for x in indices]
 
     def _generate(self, declarations=None):
         if self.backend == 'opencl':
@@ -887,14 +935,13 @@ class ScanBase(object):
     def _append_cython_arg_data(self, all_py_data, all_c_data, py_data,
                                 c_data):
         if len(c_data) > 0:
-            n_ignore = self._num_ignore_args(c_data)
-            all_py_data[0].extend(py_data[0][n_ignore:])
-            all_py_data[1].extend(py_data[1][n_ignore:])
-            all_c_data[0].extend(c_data[0][n_ignore:])
-            all_c_data[1].extend(c_data[1][n_ignore:])
+            select = self._not_ignored(c_data[1])
+            all_py_data[0].extend(self._filter_ignored(py_data[0], select))
+            all_py_data[1].extend(self._filter_ignored(py_data[1], select))
+            all_c_data[0].extend(self._filter_ignored(c_data[0], select))
+            all_c_data[1].extend(self._filter_ignored(c_data[1], select))
 
     def _generate_cython_code(self, declarations=None):
-        name = self.name
         all_py_data = [[], []]
         all_c_data = [[], []]
 
@@ -910,7 +957,9 @@ class ScanBase(object):
         # Process segment function
         use_segment = True if self.is_segment_func is not None else False
         py_data, c_data, segment_expr = self._wrap_cython_code(
-            self.is_segment_func, func_type='segment', declarations=declarations)
+            self.is_segment_func, func_type='segment',
+            declarations=declarations
+        )
         self._append_cython_arg_data(all_py_data, all_c_data, py_data, c_data)
 
         # Process output expression
@@ -962,8 +1011,10 @@ class ScanBase(object):
             is_segment_start_expr=segment_expr,
             complex_map=self.complex_map
         )
+        self.source = self.tp.get_code()
         self.tp.add_code(src)
         self.tp.compile()
+        self.all_source = self.tp.source
         return getattr(self.tp.mod, 'py_' + self.name)
 
     def _wrap_ocl_function(self, func, func_type=None, declarations=None):
@@ -977,9 +1028,9 @@ class ScanBase(object):
                 args=', '.join(c_data[1])
             )
 
-            n_ignore = self._num_ignore_args(c_data)
-            arguments = c_data[0][n_ignore:]
-            c_args = c_data[1][n_ignore:]
+            select = self._not_ignored(c_data[1])
+            arguments = self._filter_ignored(c_data[0], select)
+            c_args = self._filter_ignored(c_data[1], select)
         else:
             if func_type is 'input':
                 if self.backend == 'opencl':
@@ -1052,6 +1103,18 @@ class ScanBase(object):
             is_segment_start_expr=segment_expr,
             preamble=preamble
         )
+        self.source = preamble
+        if knl.first_level_scan_info.kernel.program.source:
+            self.all_source = '\n'.join([
+                '// ----- Level 1 ------',
+                knl.first_level_scan_info.kernel.program.source,
+                '// ----- Level 2 ------',
+                knl.second_level_scan_info.kernel.program.source,
+                '// ----- Final output ------',
+                knl.final_update_info.kernel.program.source,
+            ])
+        else:
+            self.all_source = self.source
         return knl
 
     def _generate_cuda_kernel(self, declarations=None):
@@ -1072,6 +1135,9 @@ class ScanBase(object):
             is_segment_start_expr=segment_expr,
             preamble=preamble
         )
+        self.source = preamble
+        # FIXME: Difficult to get the pycuda sources
+        self.all_source = self.source
         return knl
 
     def _add_address_space(self, arg):
@@ -1108,9 +1174,17 @@ class ScanBase(object):
         else:
             return np.asarray(x)
 
+    @profile
     def __call__(self, **kwargs):
         c_args_dict = {k: self._massage_arg(x) for k, x in kwargs.items()}
-        output_arg_keys = self.output_func.arg_keys[self._get_backend_key()]
+        if self._get_backend_key() in self.output_func.arg_keys:
+            output_arg_keys = self.output_func.arg_keys[
+                self._get_backend_key()
+            ]
+        else:
+            raise ValueError("No kernel arguments found for backend = %s, "
+                             "use_openmp = %s, use_double = %s" %
+                             self._get_backend_key())
 
         if self.backend == 'cython':
             size = len(c_args_dict[output_arg_keys[1]])
@@ -1158,6 +1232,9 @@ class Scan(object):
                                 complex_map=complex_map,
                                 backend=backend)
 
+    def __dir__(self):
+        return sorted(dir(self.scan) + ['scan'])
+
     def __getattr__(self, name):
         return getattr(self.scan, name)
 
