@@ -6,7 +6,7 @@ from pytest import importorskip
 
 from ..config import get_config, use_config
 from ..array import wrap, zeros
-from ..types import annotate
+from ..types import annotate, declare
 from ..parallel import Elementwise, Reduction, Scan
 from ..low_level import atomic_inc
 from .test_jit import g
@@ -99,10 +99,6 @@ class ParallelUtilsBase(object):
         with use_config(use_openmp=True):
             self._test_large_scan(backend='cython')
 
-    def test_large_scan_works_cython_parallel(self):
-        with use_config(use_openmp=True):
-            self._test_large_scan(backend='cython')
-
     def test_scan_works_opencl(self):
         importorskip('pyopencl')
         self._test_scan(backend='opencl')
@@ -195,6 +191,20 @@ class ParallelUtilsBase(object):
         importorskip('pycuda')
         self._test_atomic_inc(backend='cuda')
 
+    def test_repeated_scans_with_different_settings(self):
+        importorskip('pyopencl')
+        with use_config(use_double=False):
+            self._test_unique_scan(backend='opencl')
+
+        with use_config(use_double=True):
+            self._test_unique_scan(backend='opencl')
+
+        with use_config(use_openmp=False):
+            self._test_unique_scan(backend='cython')
+
+        with use_config(use_openmp=True):
+            self._test_unique_scan(backend='cython')
+
 
 class TestParallelUtils(ParallelUtilsBase, unittest.TestCase):
     def setUp(self):
@@ -224,6 +234,9 @@ class TestParallelUtils(ParallelUtilsBase, unittest.TestCase):
         # Then
         y.pull()
         self.assertTrue(np.allclose(y.data, a * np.sin(x.data) + b))
+        self.assertTrue(len(e.source) > 100)
+        self.assertTrue(len(e.all_source) > 100)
+        self.assertTrue(len(e.all_source) >= len(e.source))
 
     def _check_elementwise_with_constant(self, backend):
         # Given
@@ -253,6 +266,10 @@ class TestParallelUtils(ParallelUtilsBase, unittest.TestCase):
         # Then
         self.assertAlmostEqual(result, 0.5, 6)
 
+        self.assertTrue(len(r.source) > 100)
+        self.assertTrue(len(r.all_source) > 100)
+        self.assertTrue(len(r.all_source) >= len(r.source))
+
     def _check_reduction_min(self, backend):
         x = np.linspace(0, 1, 1000) / 1000
         x = wrap(x, backend=backend)
@@ -325,6 +342,10 @@ class TestParallelUtils(ParallelUtilsBase, unittest.TestCase):
         # Then
         np.testing.assert_equal(expect, result)
 
+        self.assertTrue(len(scan.source) > 100)
+        self.assertTrue(len(scan.all_source) > 100)
+        self.assertTrue(len(scan.all_source) >= len(scan.source))
+
     def _test_large_scan(self, backend):
         # Given
         a = np.ones(3000000, dtype=np.int32)
@@ -405,7 +426,7 @@ class TestParallelUtils(ParallelUtilsBase, unittest.TestCase):
 
         @annotate(int='i, prev_item, item, N', ary='intp',
                   unique='intp', unique_count='intp')
-        def output_f(i, prev_item, item, N, ary, unique, unique_count):
+        def output_f(i, N, ary, unique, unique_count, item, prev_item):
             if item != prev_item:
                 unique[item - 1] = ary[i]
             if i == N - 1:
@@ -529,6 +550,10 @@ class TestParallelUtilsJIT(ParallelUtilsBase, unittest.TestCase):
         y.pull()
         self.assertTrue(np.allclose(y.data, a * np.sin(x.data) + b))
 
+        self.assertTrue(len(e.source) > 100, e.source)
+        self.assertTrue(len(e.all_source) > 100, e.all_source)
+        self.assertTrue(len(e.all_source) >= len(e.source))
+
     def _check_elementwise_with_constant(self, backend):
         # Given
         @annotate
@@ -557,6 +582,10 @@ class TestParallelUtilsJIT(ParallelUtilsBase, unittest.TestCase):
         # Then
         self.assertAlmostEqual(result, 0.5, 6)
 
+        self.assertTrue(len(r.source) > 100)
+        self.assertTrue(len(r.all_source) > 100, r.all_source)
+        self.assertTrue(len(r.all_source) >= len(r.source))
+
     def _check_reduction_min(self, backend):
         x = np.linspace(0, 1, 1000) / 1000
         x = wrap(x, backend=backend)
@@ -630,6 +659,9 @@ class TestParallelUtilsJIT(ParallelUtilsBase, unittest.TestCase):
 
         # Then
         np.testing.assert_equal(expect, result)
+        self.assertTrue(len(scan.source) > 100, scan.source)
+        self.assertTrue(len(scan.all_source) > 100, scan.all_source)
+        self.assertTrue(len(scan.all_source) >= len(scan.source))
 
     def _test_large_scan(self, backend):
         # Given
